#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <wchar.h>
#include <string>
#include <algorithm>
#include <functional>
#include "plugin.hpp"

#include "types.hpp"
#include "tools.hpp"

extern PluginStartupInfo startupInfo;
extern const GUID MainGuid;

void errorMessage(const wchar_t *msg)
{
  const wchar_t *msgItems[]={L"Error", L"", L"Ok"};
  msgItems[1] = msg;
  startupInfo.Message(&MainGuid,
                      &__uuidof(struct __declspec(uuid("{D9640061-13DF-4CAD-B144-8DDDB243843A}")) TMsgGuid),
                      FMSG_WARNING, NULL,
                      msgItems,
                      std::size(msgItems), 1);
}

intptr_t messageBox(unsigned long flags, const wchar_t * const *items, int noItems, int noButtons)
{
  return startupInfo.Message(&MainGuid,
                             &__uuidof(struct __declspec(uuid("{2FBC7172-7605-4285-971D-E1EA9A00E585}")) TMsgGuid),
                              flags, NULL, items,
                              noItems, noButtons);
}

const wchar_t* pointToName(const wchar_t *path)
{
  const wchar_t *namePtr = path;
  while(*path)
  {
    if(*path == L'\\' || *path == L'/' || *path==L':') namePtr = path+1;
    path++;
  }
  return namePtr;
}

wchar_t* pointToName(wchar_t *path)
{
    wchar_t *namePtr = path;
    while(*path)
    {
        if(*path == L'\\' || *path == L'/' || *path == L':') namePtr = path + 1;
        path++;
    }
    return namePtr;
}

const wchar_t* pointToExt(const wchar_t *path)
{
    int i = lstrlenW(path);
    auto ptr = path + i;
    while(*ptr != L'.' && i != 0) { ptr--; i--; }
    if(i)
        return ptr + 1;

    return nullptr;
}

wchar_t* pointToExt(wchar_t *path)
{
    int i = lstrlenW(path);
    auto ptr = path + i;
    while(*ptr != L'.' && i != 0) { ptr--; i--; }
    if(i)
        return ptr + 1;

    return nullptr;
}

char* trim(char *str)
{
    int i, j, len;

    len = lstrlen(str);

    for(i = len - 1; i >= 0; i--)
        if(str[i] != ' ') break;
    str[i + 1] = 0;

    for(i = 0; i < len; i++)
        if(str[i] != ' ') break;
    if(i == 0) return str;

    for(j = i; j < len; j++)
        str[j - i] = str[j];

    return str;
}

wchar_t* trim(wchar_t *str)
{
  int i,j,len;

  len = lstrlenW(str);

  for(i = len-1; i>=0; i--)
    if(str[i] != L' ') break;
  str[i+1] = 0;

  for(i = 0; i < len; i++)
    if(str[i] != L' ') break;
  if(i == 0) return str;

  for(j = i; j < len; j++)
    str[j-i] = str[j];

  return str;
}

wchar_t* truncPathStr(wchar_t* newName, const wchar_t* name, int noChars)
{
  int size = lstrlenW(name);
  if(size > noChars)
  {
    if(name[1] == L':' && name[2] == L'\\')
    {
      lstrcpynW(newName, name, 4);
      lstrcatW(newName, L"...");
      lstrcatW(newName, name+size-noChars+6);
    }
    else
    {
      lstrcpyW(newName, L"...");
      lstrcatW(newName, name+size-noChars+3);
    }
  }
  else
    lstrcpyW(newName, name);
  return newName;
}

void addEndSlash(wchar_t *path)
{
  int length=lstrlenW(path);
  if(length == 0 || path[length-1] != L'\\') lstrcatW(path, L"\\");
}

void addEndSlash(std::wstring &path)
{
    if(path.empty() || path.back() != L'\\')
    {
        path.append(L"\\");
    }
}

void quoteSpaceOnly(wchar_t *str)
{
  wchar_t *tmp = str;
  wchar_t to[300];
  int len = lstrlenW(str);

  wmemcpy(to, str, len);
  *(to + len) = 0;

  while(*tmp)
  {
    if(*tmp == ' ')
    {
      *to = '"';
      wmemcpy(to + 1, str, len);
      *(to + len + 1) = '"';
      *(to + len + 2) = 0;
      break;
    }
    ++tmp;
  }

  lstrcpyW(str, to);
}

static constexpr BYTE invalidChars[] = { '\\', '/', ':', '?', '*', '<', '>', '|' };

bool isValidChar(BYTE ch)
{
  if(ch < 0x20 || ch > 0x7F) return false;
  if(std::any_of(std::cbegin(invalidChars), std::cend(invalidChars), [ch](char x) { return x == ch; }))
      return false;

  return true;
}

bool isValidFolderChar(BYTE ch)
{
  if(ch >= 0x80 && ch <= 0xAF) return true;
  if(ch >= 0xE0 && ch <= 0xF1) return true;
  if(ch < 0x20 || ch > 0x7F) return false;
  if(std::any_of(std::cbegin(invalidChars), std::cend(invalidChars), [ch](char x) { return x == ch; }))
      return false;
  return true;
}

bool isEnAlphaNum(BYTE ch)
{
  if(ch >= '0' && ch <= '9') return true;
  if(ch >= 'A' && ch <= 'Z') return true;
  if(ch >= 'a' && ch <= 'z') return true;
  return false;
}

void makeTrDosName(wchar_t* dest, const SCLHdr &hdr, bool ShowExt, int width)
{
    std::transform(std::cbegin(hdr.name), std::cend(hdr.name), dest, Cp866ToUcs2);

    BYTE ch1, ch2, ch3;
    ch1 = hdr.type;
    ch2 = hdr.start & 0xff;
    ch3 = hdr.start >> 8;

    if(ShowExt &&
        isEnAlphaNum(ch1) &&
        isEnAlphaNum(ch2) &&
        isEnAlphaNum(ch3))
    {
        dest[width - 3] = ch1;
        dest[width - 2] = ch2;
        dest[width - 1] = ch3;
    }
    else
    {
        dest[width - 3] = L'<';
        dest[width - 2] = ch1 ? ch1 : ' ';
        dest[width - 1] = L'>';
    }
    dest[width] = 0;
}

WORD crc(BYTE* ptr, size_t len)
{
  WORD crc = 0;
  
  for(; len; len--)
  {
    BYTE tmp = (crc ^ *ptr++) & 0xFF;
    
    WORD bCRC = 0;
    for(int i = 0; i < 8; ++i)
    {
      WORD lBit = bCRC & 1;
      bCRC >>= 1; bCRC |= lBit<<15; // RRA crc

      if((tmp & 1) ^ lBit) bCRC ^= 0xA001;
      tmp >>= 1;
    }
    crc = (bCRC ^ (crc<<8 | crc>>8)) & 0xFFFF;
  }
  return (crc<<8 | crc>>8) & 0xFFFF;
}

DWORD calculateCheckSum(BYTE* ptr, size_t size)
{
  DWORD sum = 0;
  while(size--) sum += *ptr++;
  return sum;
}

WORD calculateCheckSum(HoHdr& hdr)
{
  WORD sum = 0;
  BYTE* p = (BYTE*)&hdr;
  for(int i = 0; i < sizeof(HoHdr)-2; ++i) sum += *p++;
  return (257*sum + 105) & 0xFFFF;
}

char* make8x3name(const wchar_t* source, char* dest)
{
  const wchar_t* ptr = source;
  //  
  for(int i = 0; i < 8; ++i)
  {
    if(!*ptr || *ptr == L'.') break;

    //   cp866
    WideCharToMultiByte(866, 0, ptr++, 1, &dest[i], 1, nullptr, nullptr);
  }
  //     -  
  while(*ptr && *ptr != L'.') ++ptr;

  if(*ptr == L'.')
  {
    ++ptr;
    //  
    for(int i = 0; i < 3; ++i)
    {
      if(!*ptr) break;

      //   cp866
      WideCharToMultiByte(866, 0, ptr++, 1, &dest[8 + i], 1, nullptr, nullptr);
    }
  }
  return dest;
}

const wchar_t *GetMsg(intptr_t MsgId)
{
  return startupInfo.GetMsg(&MainGuid, MsgId);
};

void SafeDlgFree(HANDLE Dlg)
{
    if(Dlg != INVALID_HANDLE_VALUE)
    {
        startupInfo.DialogFree(Dlg);
    }
}

wchar_t Cp866ToUcs2(char x)
{
    wchar_t y;
    if(x)
    {
        //   cp866
        MultiByteToWideChar(866, 0, &x, 1, &y, 1);
    }
    else
    {
        y = L' ';
    }

    return y;
}

FileType detectFileType(HANDLE file)
{
    DWORD fileSize = GetFileSize(file, NULL);

    DWORD noBytesRead;
    BYTE  buf[sizeof(HoHdr)];
    ReadFile(file, buf, sizeof(HoHdr), &noBytesRead, 0);
    SetFilePointer(file, 0, NULL, FILE_BEGIN);

    //    HoBeta  
    HoHdr* hdr = (HoHdr*)buf;
    if(hdr->checkSum == calculateCheckSum(*hdr) &&
        fileSize >= sectorSize * hdr->noSecs + sizeof(HoHdr)) return FMT_HOBETA;

    //    SCL  

    //      ,
    //    scl- 
    //   SN

    char signature[] = { 'S', 'I', 'N', 'C', 'L', 'A', 'I', 'R' };
    BYTE no_files = buf[sizeof(signature)];
    if(fileSize < no_files*(sectorSize + sizeof(SCLHdr)) + 8 + 1 + 4 ||
        memcmp(buf, signature, sizeof(signature))) return FMT_PLAIN;

    DWORD totalSecs = 0;
    SetFilePointer(file, 8 + 1, NULL, FILE_BEGIN);
    for(int i = 0; i < no_files; ++i)
    {
        SCLHdr fHdr;
        ReadFile(file, &fHdr, sizeof(SCLHdr), &noBytesRead, 0);
        totalSecs += fHdr.noSecs;
    }
    SetFilePointer(file, 0, NULL, FILE_BEGIN);

    if(fileSize < totalSecs*sectorSize + no_files * sizeof(SCLHdr) + 8 + 1 + 4)
        return FMT_PLAIN;
    else
        return FMT_SCL;
}

int str2int(const wchar_t* p, int len)
{
    int result = -1;
    while(*p == ' ') ++p;

    for(int i = 0; i < len; ++i)
    {
        BYTE digit = 0xFF;
        if(*p >= '0' && *p <= '9') digit = *p - '0';

        if(digit == 0xFF) break;
        if(result == -1) result = 0;

        result = 10 * result + digit;
        p++;
    }
    return result;
}

void makePCName(const char *from, wchar_t *to)
{
    int first = 0, last = 7;
    while(first <= last)
    {
        if(isValidChar(from[first])) break;
        ++first;
    }
    while(last >= first)
    {
        if(isValidChar(from[last])) break;
        --last;
    }
    BYTE  noChars = 0;

    if(first > last)
    {
        std::fill_n(to, 8, L'_');
        noChars = 8;
    }
    else
    {
        while(first <= last)
        {
            auto ch = from[first++];
            to[noChars++] = isValidChar(ch) ? Cp866ToUcs2(ch) : L'_';
        }
    }

    //    
    if(noChars == 3 || noChars == 4)
    {
        if(!wcsnicmp(to, L"com", 3) ||
            !wcsnicmp(to, L"lpt", 3) ||
            !wcsnicmp(to, L"prn", 3) ||
            !wcsnicmp(to, L"con", 3) ||
            !wcsnicmp(to, L"aux", 3) ||
            !wcsnicmp(to, L"nul", 3)) to[noChars++] = L'_';
    }
    to[noChars] = 0;
    trim(to);
}

DWORD writeSCLHeader(HANDLE file, BYTE no_files)
{
    static const BYTE  signature[] = { 'S', 'I', 'N', 'C', 'L', 'A', 'I', 'R' };
    DWORD noBytesWritten;

    WriteFile(file, signature, sizeof(signature), &noBytesWritten, NULL);
    WriteFile(file, &no_files, 1, &noBytesWritten, NULL);

    //    
    SetFilePointer(file, no_files * sizeof(SCLHdr) + 8 + 1, NULL, FILE_BEGIN);
    SetEndOfFile(file);

    return (0x255 + no_files);
}
