#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>
#include <wchar.h>
#include <algorithm>

#include "types.hpp"
#include "types_trd.hpp"
#include "manager.hpp"
#include "detector.hpp"
#include "tools.hpp"
#include "lang.hpp"

extern Options           op;
extern Detector*         detector;

char dsSignature[] = { 'D', 'i', 'r', 'S', 'y', 's', '1', '0', '0' };

bool Manager::openHostFile(void) { return fmt->open(img); }
void Manager::closeHostFile(void){        fmt->close(img); }

bool Manager::read(int trk, int sec, BYTE* buf)
{
  bool result = fmt->read(img, trk, sec, buf);
  if(!result)
  {
    const wchar_t *msgItems[]{ GetMsg(MError), GetMsg(MCanNotReadSec), nullptr, GetMsg(MOk) };

    wchar_t msg[30];
    wsprintfW(msg, GetMsg(MTrkSec), trk, sec);
    msgItems[2] = msg;
    messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
  }
  return result;
}

bool Manager::write(int trk, int sec, BYTE* buf)
{
  bool result = fmt->write(img, trk, sec, buf);
  if(!result)
  {
    const wchar_t *msgItems[4]{ GetMsg(MError), GetMsg(MCanNotWriteSec), nullptr, GetMsg(MOk) };

    wchar_t msg[30];
    wsprintfW(msg, GetMsg(MTrkSec), trk, sec);
    msgItems[2] = msg;
    messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
  }
  return result;
}

bool Manager::readInfo(void)
{
  //       
  WIN32_FIND_DATAW data;
  HANDLE h = FindFirstFileW(hostFileName, &data);
  if(h == INVALID_HANDLE_VALUE) return false;
  FindClose(h);
  if(!op.useDS)
  {
    curFolderNum = 0;
    *curFolder   = 0;
  }
  
  if(CompareFileTime(&data.ftLastWriteTime, &lastModifed.ftLastWriteTime) != 0 ||
     data.nFileSizeLow != lastModifed.nFileSizeLow || op.reread)
  {
    op.reread    = false;
    lastModifed  = data;
    noFiles      = 0;
    noDelFiles   = 0;
    
    if(!fmt->reload(img)) return false;
    if(!openHostFile()) return false;
    
    //  0- 
    for(int i = 0; i < 16; ++i) read(0, i, zeroTrk+i*sectorSize);

    //    
    BYTE* ptr = zeroTrk;
    for(; noFiles < 142; ++noFiles)
    {
      if(*ptr == 0x00) break;
      if(*ptr == 0x01) ++noDelFiles;

      memcpy(&files[noFiles], ptr, sizeof(FileHdr));
      //  ,   
      //    MOVE
      *ptr = 0x00; 
      ptr += sizeof(FileHdr);
    }

    memcpy(&diskInfo, zeroTrk+8*sectorSize+0xE1, sizeof(DiskHdr));

    //  DirSys
    BYTE folderName[11];
    if(curFolderNum != 0)
    {
        static_assert(std::size(folderName) == _countof(folders[curFolderNum - 1]));
        std::copy(std::cbegin(folders[curFolderNum - 1]), std::cend(folders[curFolderNum - 1]), folderName);
    }

    noFolders    = 0;
    noDelFolders = 0;

    dsOk  = false;
    ptr   = zeroTrk+9*sectorSize;
    if(!memcmp(ptr+2, dsSignature, sizeof(dsSignature)))
    {
      dsCRC = *(WORD*)ptr;
      memcpy(fileMap,   ptr+2+sizeof(dsSignature),     128);
      memcpy(folderMap, ptr+2+sizeof(dsSignature)+128, 127);
      ptr = zeroTrk+10*sectorSize+11;
      for(; noFolders < 127; ++noFolders)
      {
        if(*ptr == 0x00) break;
        if(*ptr == 0x01) ++noDelFolders;

        std::copy_n(ptr, std::size(folders[noFolders]), folders[noFolders]);
        ptr += std::size(folders[noFolders]);
      }
      dsOk = (dsCRC == calcDScrc());
      //  ,   
      //    MOVE
      ptr = zeroTrk+10*sectorSize+11;
      for(int i = 0; i < noFolders; ++i) *(ptr + 11*i) = 0;
      if(!dsOk)
      {
        curFolderNum = 0;
        *curFolder   = 0;
      }
      if(curFolderNum != 0)
      {
        int i = curFolderNum > noFolders ? noFolders : curFolderNum;
        for(; i > 0; --i)
          if(!memcmp(folders[i-1], folderName, 11)) break;
        if(i == 0)
        {
          curFolderNum = 0;
          *curFolder   = 0;
        }
        else
          curFolderNum = i;
      }
    }
    makePCNames();
    closeHostFile();
  }
  return true;
}

void Manager::writeInfo(void)
{
  memcpy(zeroTrk, files, noFiles*sizeof(FileHdr));
  memcpy(zeroTrk+8*sectorSize+0xE1, &diskInfo, sizeof(DiskHdr));
  for(int i = 0; i < 9; ++i) write(0, i, zeroTrk+i*sectorSize);

  BYTE* dsPtr = zeroTrk+9*sectorSize;
  if(dsOk)
  {
    memcpy(dsPtr+2,      dsSignature, sizeof(dsSignature));

    memcpy(dsPtr+11,     fileMap,     128);
    memcpy(dsPtr+11+128, folderMap,   127);
    memcpy(dsPtr+11+256, folders,     11*noFolders);

    dsPtr[11+255]              = 0;
    dsPtr[11+256+11*noFolders] = 0;

    dsCRC = calcDScrc();
    memcpy(dsPtr, &dsCRC, 2);
    for(int i = 9; i < 16; ++i) write(0, i, zeroTrk+i*sectorSize);
  }
}

void Manager::makePCNames(void)
{
  for(int fNum = 0; fNum < noFiles; ++fNum)
  {
    const char* from = files[fNum].name;
    wchar_t* const to = pcFiles[fNum].name;

    makePCName(from, to);
    BYTE noChars = lstrlenW(to);

    BYTE dotPos = noChars;
    to[noChars++] = L'.';
    to[noChars++] = L'$';
    if(isValidChar(files[fNum].type)) to[noChars++] = Cp866ToUcs2(files[fNum].type);
    to[noChars] = 0;
    
    BYTE typeNum = 0xFF;
    if(op.detectFormat)
    {
      BYTE secs[16 * sectorSize]{};

      int noSecs = files[fNum].noSecs;
      if(noSecs > 16) noSecs = 16;
      int trk = files[fNum].trk;
      int sec = files[fNum].sec;
      int s = 0;

      for(; s < noSecs; ++s)
      {
        if(!fmt->read(img, trk, sec, secs+s*sectorSize)) break;
        ++sec;
        if(sec == 16) { ++trk; sec = 0; }
      }
      typeNum = detector->detect(files[fNum], secs, sectorSize*s, pcFiles[fNum].comment);
    }

    pcFiles[fNum].type = typeNum;

    pcFiles[fNum].skipHeader = false;
   
    if(typeNum != 0xFF)
    {
      detector->specialChar(typeNum, to+dotPos+1);
      detector->getType(typeNum, to+dotPos+2);
      pcFiles[fNum].skipHeader = detector->getSkipHeader(typeNum);
    }
    //   zxzip 
    if(fNum != 0 &&
       !memcmp(from, "********ZIP", 11) &&
       !memcmp(&files[fNum-1].type, "ZIP", 3))
    {
      if(memcmp(files[fNum-1].name, "********", 8))
        lstrcpyW(to, pcFiles[fNum-1].name);
      else
        lstrcpynW(to, pcFiles[fNum-1].name, lstrlenW(pcFiles[fNum-1].name));
    }
    //   
    int i = fNum;
    while(i-- > 0)
    {
      int len = lstrlenW(to);
      if(len && !wcsnicmp(to, pcFiles[i].name, len))
      {
        wchar_t ch = pcFiles[i].name[len-1];
        ch = (ch != 0) ? ch+1 : '0';

        //     :
        if(ch == '9'+1) ch = 'A'; 
        if(ch == 'Z'+1) ch = 'a';
        if(ch == 'z'+1) ch = '0';

        to[len-1] = ch; to[len] = 0;
        break;
      }
    }
  }
  //   
  for(int fNum = 0; fNum < noFolders; ++fNum)
  {
    wchar_t* ptr = pcFolders[fNum];

    char FolderName[9];
    std::copy_n(folders[fNum], 8, FolderName);
    FolderName[8] = 0;
    trim(FolderName);
    int n = lstrlen(FolderName);
    for(int i = 0; i < n; ++i)
    {
        if(!isValidFolderChar(FolderName[i]))
        {
            FolderName[i] = '_';
        }
    }

    std::transform(FolderName, FolderName + n, ptr, Cp866ToUcs2);
    ptr[n] = 0;

    char FolderExt[4];
    std::copy_n(folders[fNum] + 8, 3, FolderExt);
    FolderExt[3] = 0;
    trim(FolderExt);
    n = lstrlen(FolderExt);
    for(int i = 0; i < n; ++i)
    {
        if(!isValidFolderChar(FolderExt[i]))
        {
            FolderExt[i] = '_';
        }
    }

    wchar_t ext[4];
    std::transform(FolderExt, FolderExt + n, ext, Cp866ToUcs2);
    ext[n] = 0;
    if(ext[0] != 0)
    {
      lstrcatW(ptr, L".");
      lstrcatW(ptr, ext);
    }
    int i = fNum;
    while(i-- > 0)
    {
      if(folderMap[i] != folderMap[fNum]) continue;
      
      int len = lstrlenW(ptr);
      if(!wcsnicmp(ptr, pcFolders[i], len))
      {
        if(lstrlenW(pcFolders[i]) == len)
        {
          lstrcatW(ptr, L".0");
          break;
        }
        wchar_t ch = pcFolders[i][len+1];
        if(pcFolders[i][len] == L'.' && isEnAlphaNum(ch))
        {
          lstrcatW(ptr, L".0");
          ++ch;
          if(ch == L'9'+1) ch = L'A'; 
          if(ch == L'Z'+1) ch = L'a';
          if(ch == L'z'+1) ch = L'0';
          pcFolders[fNum][len+1] = ch;
          break; 
        }
      }
    }
  }
}

WORD Manager::calcDScrc(void)
{
  return crc(zeroTrk+9*sectorSize+2, 256+9+11*noFolders);
}
